class BracketError(Exception):
    """ Exception raised for errors with brackets input by the user.

    Attributes:
        message -- explanation of the error.
    """

    def __init__(self, message="Brackets are not in the proper format."):
        self.message = message
        super().__init__(self.message)



import milesianpy.common_operators as common_operators
import milesianpy.error_handling.operator_error as operator_error
import milesianpy.error_handling.bracket_error as bracket_error
import milesianpy.error_handling.decimal_error as decimal_error
import milesianpy.error_handling.invalid_character_error as invalid_character_error


class BasicParser:
    @staticmethod
    def bracket_multiplication_insertion(user_input: list):
        previous_value = ''
        new_user_input = []
        operator_error_class = operator_error.OperatorError

        for x in user_input:
            if previous_value.isnumeric() and x in common_operators.CommonOperators.LEFT_BRACKET.value:
                new_user_input.append('*')
                new_user_input.append(x)

            elif previous_value in common_operators.CommonOperators.RIGHT_BRACKET.value and x.isnumeric():
                raise operator_error_class("Missing operator.")

            else:
                previous_value = x
                new_user_input.append(x)

        return new_user_input

    @staticmethod
    def check_bracket_count(user_input: list):
        left_bracket_count = 0
        right_bracket_count = 0
        bracket_error_class = bracket_error.BracketError

        for x in common_operators.CommonOperators.LEFT_BRACKET.value:
            left_bracket_count += user_input.count(x)

        for x in common_operators.CommonOperators.RIGHT_BRACKET.value:
            right_bracket_count += user_input.count(x)

        if right_bracket_count != left_bracket_count:
            raise bracket_error_class("No closing bracket.")

        if left_bracket_count > 0:
            return True

        return False

    @staticmethod
    def check_if_empty_bracket(user_input: list):
        left_value = ''
        bracket_error_class = bracket_error.BracketError

        for x in user_input:
            if x in common_operators.CommonOperators.LEFT_BRACKET.value:
                left_value = x

            elif x in common_operators.CommonOperators.RIGHT_BRACKET.value and left_value:
                raise bracket_error_class("No content in bracket.")

            else:
                left_value = ''

        return False

    @staticmethod
    def check_double_operators(user_input: list, passed_valid_operators: list = None):
        previous_value = ''
        operator_error_class = operator_error.OperatorError

        if passed_valid_operators is None:
            passed_valid_operators = common_operators.CommonOperators.VALID_OPERATORS.value

        for x in user_input:
            if previous_value is x in passed_valid_operators:
                raise operator_error_class(previous_value + " cannot be beside " + x + ".")

            else:
                previous_value = x

        return False

    @staticmethod
    def check_decimals(user_input: list):
        decimal = '.'
        previous_value = ''
        decimal_error_class = decimal_error.DecimalError

        for x in user_input:
            if x is decimal and previous_value is not decimal:
                if previous_value == '' or previous_value in common_operators.CommonOperators.VALID_CHARACTERS.value:
                    raise decimal_error_class("Decimals require numbers on both sides.")

                previous_value = decimal

            elif x is decimal and previous_value is decimal:
                raise decimal_error_class("Too many decimals for one number.")

            else:
                if previous_value is decimal and x in common_operators.CommonOperators.VALID_CHARACTERS.value:
                    raise decimal_error_class("Decimals require numbers on both sides.")

                previous_value = x

        if user_input[len(user_input) - 1] is '.':
            raise decimal_error_class("Decimals require numbers on both sides.")

        return True

    @staticmethod
    def check_values(user_input: list, passed_valid_characters: list = None):
        invalid_character_error_class = invalid_character_error.InvalidCharacterError

        if passed_valid_characters is None:
            passed_valid_characters = common_operators.CommonOperators.VALID_CHARACTERS.value

        for x in user_input:
            if x not in passed_valid_characters and not x.isalnum():
                raise invalid_character_error_class("Invalid character: " + x + ".")

        return True

    @staticmethod
    def standardize_brackets(user_input: list):
        user_input = list(map(lambda x: x.replace('{', '('), user_input))
        user_input = list(map(lambda x: x.replace('}', ')'), user_input))
        user_input = list(map(lambda x: x.replace('[', '('), user_input))
        user_input = list(map(lambda x: x.replace(']', ')'), user_input))

        return user_input



import milesianpy.common_operators as common_operators


class NumberParser:
    @staticmethod
    def convert_to_nums(user_input: list):
        converted_equation = []
        valid_num = ""

        for x in user_input:
            if x in common_operators.CommonOperators.VALID_CHARACTERS_NO_DECIMALS.value:
                if valid_num:
                    converted_equation.append(float(valid_num))
                    valid_num = ""

                converted_equation.append(x)

            elif x.isnumeric():
                valid_num += x

            elif x == ".":
                valid_num += x

        if valid_num:
            converted_equation.append(float(valid_num))

        converted_equation = NumberParser.is_negative(converted_equation)

        return converted_equation

    @classmethod
    def is_negative(cls, user_input: list):
        iterator = 0

        while len(user_input) > iterator:
            if user_input[iterator] == '-':
                if type(user_input[iterator - 1]) is not float and type(user_input[iterator + 1]) is float:
                    user_input[iterator + 1] = user_input[iterator + 1] * -1
                    del user_input[iterator]

            iterator += 1

        return user_input














import pytest


def test_runner(setup_data, function_name, passed_function):
    for x in setup_data[function_name]:
        equation = x['test_data']
        if x['test_result']['exception'] is False:
            response = passed_function(equation)
            assert response == x['test_result']['result']

        elif x['test_result']['exception'] is True:
            with pytest.raises(Exception) as error:
                passed_function(equation)
            assert str(error.value) == x['test_result']['result']



class OperatorParser:
    @staticmethod
    def addition(equation: list):
        return equation[0] + equation[1]

    @staticmethod
    def subtraction(equation: list):
        return equation[0] - equation[1]

    @staticmethod
    def multiplication(equation: list):
        return equation[0] * equation[1]

    @staticmethod
    def division(equation: list):
        return equation[0] / equation[1]

    @staticmethod
    def exponent(equation: list):
        index = 0
        outcome = equation[0]

        while equation[1] - 1 > index:
            outcome *= equation[0]
            index += 1

        return outcome



import milesianpy.parsers.operator_parser as operator_parser
import milesianpy.parsers.number_parser as number_parser
import milesianpy.parsers.basic_parser as basic_parser
import milesianpy.common_operators as common_operators


class NoVariableCalculation:
    @staticmethod
    def no_variable_basic_calculation(user_input: list):
        number_class = number_parser.NumberParser()
        basic_class = basic_parser.BasicParser()

        has_brackets = basic_class.check_bracket_count(user_input)

        if has_brackets:
            basic_class = basic_parser.BasicParser()
            user_input = basic_class.standardize_brackets(user_input)
            user_input = basic_class.bracket_multiplication_insertion(user_input)
            user_input = number_class.convert_to_nums(user_input)

            right_bracket_index = user_input.index(')')
            left_bracket_index = NoVariableCalculation.final_instance(user_input, '(')

            if right_bracket_index < left_bracket_index:
                response = NoVariableCalculation.multiple_nested_brackets(user_input)

            else:
                response = NoVariableCalculation.singular_nested_brackets(user_input)
                response = NoVariableCalculation.calculate_values(user_input)

        else:
            user_input = number_class.convert_to_nums(user_input)
            response = NoVariableCalculation.calculate_values(user_input)

        return response

    @classmethod
    def multiple_nested_brackets(cls, user_input: list):
        left_bracket_count = user_input.count('(')

        while left_bracket_count > 0:
            right_bracket_index = user_input.index(')')

            spliced_list = user_input[:right_bracket_index + 1]
            left_bracket_index = NoVariableCalculation.final_instance(spliced_list, '(')
            spliced_list = spliced_list[left_bracket_index:]

            calculated_value = NoVariableCalculation.singular_nested_brackets(spliced_list)
            user_input[right_bracket_index] = calculated_value
            del user_input[left_bracket_index: right_bracket_index]
            left_bracket_count = user_input.count('(')

        user_input = NoVariableCalculation.calculate_values(user_input)
        return user_input

    @classmethod
    def singular_nested_brackets(cls, user_input: list):
        left_bracket_count = user_input.count('(')
        calculated_value = 0

        while left_bracket_count > 0:
            holder = NoVariableCalculation.bracket_recursion(user_input)
            left_bracket_count_holder = holder.count('(')

            if left_bracket_count_holder == 0:
                calculated_value = NoVariableCalculation.calculate_values(holder)

                left_value = NoVariableCalculation.final_instance(user_input, '(')

                right_value = user_input.index(')')

                del user_input[left_value + 1:right_value + 1]
                user_input[left_value] = calculated_value

            left_bracket_count = user_input.count('(')

        return calculated_value

    @classmethod
    def calculate_values(cls, calculate_values: list):
        iterator = 0
        calculated_sum = calculate_values[0]
        operator = operator_parser.OperatorParser()

        while len(calculate_values) > iterator:
            if calculate_values[iterator] in common_operators.CommonOperators.VALID_OPERATORS.value:
                if calculate_values[iterator] == '+':
                    calculated_sum = operator.addition([calculated_sum, calculate_values[iterator + 1]])

                elif calculate_values[iterator] == '-':
                    calculated_sum = operator.subtraction([calculated_sum, calculate_values[iterator + 1]])

                elif calculate_values[iterator] == '*':
                    calculated_sum = operator.multiplication([calculated_sum, calculate_values[iterator + 1]])

                elif calculate_values[iterator] == '/':
                    calculated_sum = operator.division([calculated_sum, calculate_values[iterator + 1]])

                elif calculate_values[iterator] == '^':
                    calculated_sum = operator.exponent([calculated_sum, calculate_values[iterator + 1]])

                iterator += 1

            else:
                iterator += 1

        return calculated_sum

    @classmethod
    def bracket_recursion(cls, user_input: list):
        index = NoVariableCalculation.final_instance(user_input, '(')
        left_bracket_index = index
        right_bracket_index = user_input.index(')')

        return user_input[left_bracket_index + 1: right_bracket_index]

    @classmethod
    def final_instance(cls, passed_list: list, value: str):
        index = 0
        for i in range(len(passed_list) - 1, 0, -1):
            if passed_list[i] == value:
                index = i
                break

        return index
